/* #define DEBUG */
/* #define VERBOSE_DEBUG */

#include <linux/kernel.h>
#include <linux/utsname.h>
#include <linux/module.h>

#include "u_ether.h"


#define DRIVER_DESC		"Gadget Tactical modem"
#define DRIVER_VERSION		"Gadget Tactical modem V1.1"
                                                   
/*-------------------------------------------------------------------------*/

/* DO NOT REUSE THESE IDs with a protocol-incompatible driver!!  Ever!!
 * Instead:  allocate your own, using normal USB-IF procedures.
 */

/* Thanks to NetChip Technologies for donating this product ID.
 * It's for devices with only this composite CDC configuration.
 */
#define TAC_VENDOR_ID		0x1e2d	/* Cinterion */
#define TAC_PRODUCT_ID		0x0053	/* PHS8 module */

/*-------------------------------------------------------------------------*/

/*
 * Kbuild is not very cooperative with respect to linking separately
 * compiled library objects into one module.  So for now we won't use
 * separate compilation ... ensuring init/exit sections work to shrink
 * the runtime footprint, and giving us at least some parts of what
 * a "gcc --combine ... part1.c part2.c part3.c ... " build would.
 */

#include "composite.c"
#include "usbstring.c"
#include "config.c"
#include "epautoconf.c"

#include "u_serial.c"
#include "f_serial.c"
#include "f_acm.c"
#include "f_ecm.c"
#include "u_ether.c"

/*-------------------------------------------------------------------------*/

static struct usb_device_descriptor device_desc = {
	.bLength =		sizeof device_desc,
	.bDescriptorType =	USB_DT_DEVICE,

	.bcdUSB =		cpu_to_le16(0x0200),

    //USB_CLASS_COMM ,,, USB_CLASS_VENDOR_SPEC
	.bDeviceClass =		USB_CLASS_COMM,
	.bDeviceSubClass =	0,
	.bDeviceProtocol =	0,
	/* .bMaxPacketSize0 = f(hardware) */

	/* Vendor and product id can be overridden by module parameters.  */
	.idVendor =		cpu_to_le16(TAC_VENDOR_ID),
	.idProduct =		cpu_to_le16(TAC_PRODUCT_ID),
	/* .bcdDevice = f(hardware) */
	/* .iManufacturer = DYNAMIC */
	/* .iProduct = DYNAMIC */
	/* NO SERIAL NUMBER */
	.bNumConfigurations =	1,
};

static struct usb_otg_descriptor otg_descriptor = {
	.bLength =		sizeof otg_descriptor,
	.bDescriptorType =	USB_DT_OTG,

	/* REVISIT SRP-only hardware is possible, although
	 * it would not be called "OTG" ...
	 */
	.bmAttributes =		USB_OTG_SRP | USB_OTG_HNP,
};

static const struct usb_descriptor_header *otg_desc[] = {
	(struct usb_descriptor_header *) &otg_descriptor,
	NULL,
};


/* string IDs are assigned dynamically */

#define STRING_MANUFACTURER_IDX		0
#define STRING_PRODUCT_IDX		1

static char manufacturer[50];

static struct usb_string strings_dev[] = {
	[STRING_MANUFACTURER_IDX].s = manufacturer,
	[STRING_PRODUCT_IDX].s = DRIVER_DESC,
	{  } /* end of list */
};

static struct usb_gadget_strings stringtab_dev = {
	.language	= 0x0409,	/* en-us */
	.strings	= strings_dev,
};

static struct usb_gadget_strings *dev_strings[] = {
	&stringtab_dev,
	NULL,
};

static u8 hostaddr[ETH_ALEN];

static unsigned n_ports = 4;
module_param(n_ports, uint, 0);
MODULE_PARM_DESC(n_ports, "number of ports to create, default=4");


/*-------------------------------------------------------------------------*/

/*
 * We _always_ have both CDC ECM and CDC ACM functions.
 */
static int __init TAC_do_config(struct usb_configuration *c)
{
	int	status=0, i;


	printk("enter %s\n",__FUNCTION__);

	if (gadget_is_otg(c->cdev->gadget)) {
		c->descriptors = otg_desc;
		c->bmAttributes |= USB_CONFIG_ATT_WAKEUP;
	}

	status = ext_ecm_bind_config(c, hostaddr,0);
	if (status < 0)
	{
		printk("Error ext_ecm_bind_config %d\n",status);
		return status;
	}

	for(i = 0; i < n_ports && status == 0; i++)
    {
		printk("1 gser_bind_config int: %d status: %d\n",i ,status);
        //status = gser_bind_config(c, i);
		status = acm_bind_config(c, i);
		if (status < 0)
			return status;
        if(status < 0)
		{
			printk("Error gser_bind_config int: %d status: %d\n",i ,status);
            return status;
		}
		printk("2 gser_bind_config int: %d status: %d\n",i ,status);
    }


   
	return 0;
}

static struct usb_configuration tac_config_driver = {
	.label			= "Tactical modem mirror driver",
	.bConfigurationValue	= 1,
	/* .iConfiguration = DYNAMIC */
	.bmAttributes		= USB_CONFIG_ATT_SELFPOWER,
};

/*-------------------------------------------------------------------------*/

static int __init TAC_bind(struct usb_composite_dev *cdev)
{
	int			gcnum;
	struct usb_gadget	*gadget = cdev->gadget;
	int			status;

	if (!can_support_ecm(cdev->gadget)) {
		dev_err(&gadget->dev, "controller '%s' not usable\n",
			gadget->name);
		return -EINVAL;
	}
 
	/* set up network link layer */
	status = ext_gether_setup(cdev->gadget, hostaddr,0);
	if (status < 0)
		return status;

	/* set up serial link layer */
    status = gserial_setup(cdev->gadget, n_ports);
	if (status < 0)
		goto fail1;

	gcnum = usb_gadget_controller_number(gadget);
	if (gcnum >= 0)
		device_desc.bcdDevice = cpu_to_le16(0x0300 | gcnum);
	else {
		/* We assume that can_support_ecm() tells the truth;
		 * but if the controller isn't recognized at all then
		 * that assumption is a bit more likely to be wrong.
		 */
		WARNING(cdev, "controller '%s' not recognized; trying %s\n",
			gadget->name,
			tac_config_driver.label);
		device_desc.bcdDevice =
			cpu_to_le16(0x0300 | 0x0099);
	}


	/* Allocate string descriptor numbers ... note that string
	 * contents can be overridden by the composite_dev glue.
	 */

	/* device descriptor strings: manufacturer, product */
	/* snprintf(manufacturer, sizeof manufacturer, "%s %s with %s", */
	/* 	 init_utsname()->sysname, init_utsname()->release, */
	/* 	 gadget->name); */

	snprintf(manufacturer, sizeof manufacturer, "MSIL");
	status = usb_string_id(cdev);
	if (status < 0)
		goto fail0;
	strings_dev[STRING_MANUFACTURER_IDX].id = status;
	device_desc.iManufacturer = status;

	status = usb_string_id(cdev);
	if (status < 0)
		goto fail0;
	strings_dev[STRING_PRODUCT_IDX].id = status;
	device_desc.iProduct = status;

	/* register our configuration */
	status = usb_add_config(cdev, &tac_config_driver, TAC_do_config);
	if (status < 0)
		goto fail0;

	dev_info(&gadget->dev, "%s, version: " DRIVER_VERSION "\n",
		 DRIVER_DESC);

	return 0;

fail0:
    dev_err(&gadget->dev, "fail 1 '%s' \n", gadget->name);
	gserial_cleanup();
fail1:
	dev_err(&gadget->dev, "fail 0 '%s' \n", gadget->name);
	gether_cleanup();
	return status;
}

static int __exit TAC_unbind(struct usb_composite_dev *cdev)
{
    gserial_cleanup();
	gether_cleanup();
	return 0;
}

static struct usb_composite_driver TAC_driver = {
	.name		= "g_tactical",
	.dev		= &device_desc,
	.strings	= dev_strings,
	.max_speed	= USB_SPEED_HIGH,
	.unbind		= __exit_p(TAC_unbind),
};

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR("bee");
MODULE_LICENSE("GPL");

static int __init init(void)
{
	return usb_composite_probe(&TAC_driver, TAC_bind);
}
module_init(init);

static void __exit cleanup(void)
{
	usb_composite_unregister(&TAC_driver);
}
module_exit(cleanup);
